==> worktoy-0.99.70.tar.gz: worktoy-0.99.70/src/worktoy/base/_fast_space.py <==
"""FastSpace provides the namespace object class for the FastObject class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.text import monoSpace, typeMsg

try:
  from typing import Callable, Any
except ImportError:
  Callable = object

from worktoy.desc import AttriBox, DEFAULT
from worktoy.meta import BaseNamespace, OverloadEntry
from worktoy.parse import maybe


class FastSpace(BaseNamespace):
  """EZSpace provides the namespace object class for the EZData class."""

  __field_boxes__ = None
  __inner_functions__ = None

  def __init__(self, *args, **kwargs) -> None:
    """The '__init__' method initializes the namespace object."""
    BaseNamespace.__init__(self, *args, **kwargs)
    baseClasses = self.getBaseClasses()
    if len(baseClasses) > 1:
      mclsName = self.getMetaClass().__name__
      name = self.getClassName()
      spaceName = self.__class__.__name__
      baseNames = ', '.join([b.__name__ for b in baseClasses])
      e = """When trying to create a new class named: '%s' derived from 
      metaclass: '%s', the namespace object of type '%s' supports only 
      single inheritance! Nevertheless, the following base classes were
      provided: (%s)!"""
      e2 = e % (name, mclsName, spaceName, baseNames)
      raise ValueError(monoSpace(e2))

  def _getFieldBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the field boxes."""
    return maybe(self.__field_boxes__, [])

  def _getBaseBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the base field boxes."""
    baseClasses = self.getBaseClasses()
    if not baseClasses:
      return []
    baseCls = self.getBaseClasses()[0]
    baseBoxes = getattr(baseCls, '__field_boxes__', None)
    if baseBoxes is None:
      return []
    if isinstance(baseBoxes, list):
      return baseBoxes
    e = typeMsg('baseBoxes', baseBoxes, list)
    raise TypeError(e)

  def _getAllBoxes(self, ) -> list[tuple[str, AttriBox]]:
    """This method returns all the field boxes."""
    baseBoxes = self._getBaseBoxes()
    fieldBoxes = self._getFieldBoxes()
    return [*baseBoxes, *fieldBoxes]

  def _addFieldBox(self, key: str, box: AttriBox) -> None:
    """This method adds a field box to the namespace."""
    boxes = self._getFieldBoxes()
    self.__field_boxes__ = [*boxes, (key, box)]

  def __setitem__(self, key: str, value: object) -> None:
    """This method sets the key, value pair in the namespace."""
    if isinstance(value, AttriBox):
      return self._addFieldBox(key, value)
    if callable(value):
      return BaseNamespace.__setitem__(self, key, value)
    if self.isSpecialKey(key):
      return dict.__setitem__(self, key, value)
    return self.__setitem__(key, AttriBox[type(value)](DEFAULT(value)))

  @staticmethod
  def _getattrFactory(boxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__getattr__' method which automatically
    retrieves the AttriBox instances."""

    keys = [key for (key, box) in boxes]
    defGet = {key: box.getDefaultFactory() for (key, box) in boxes}

    def __getattr__(self, key: str) -> object:
      """This automatically generated '__getattr__' method retrieves the
      AttriBox instances."""
      if key in defGet:
        setattr(self, key, defGet[key](self))
      return object.__getattribute__(self, key)

    return __getattr__

  @staticmethod
  def _initFactory(attriBoxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__init__' method which automatically
    populates the AttriBox instances."""

    keys = [key for (key, box) in attriBoxes]
    defVals = {key: box.getDefaultFactory() for (key, box) in attriBoxes}
    valTypes = {key: box.getFieldClass() for (key, box) in attriBoxes}

    def __init__(self, *args, **kwargs) -> None:
      """This automatically generated '__init__' method populates the
      AttriBox instances."""
      initValues = dict()  # Creates temporary dictionary for initial values

      #  Retrieves values from positional arguments
      for (i, (key, type_)) in enumerate(valTypes.items()):
        if len(args) > i:
          if isinstance(args[i], type_):
            initValues[key] = args[i]
          else:
            e = typeMsg(key, args[i], type_)
            raise TypeError(e)

      #  Retrieves values from keyword arguments
      for (key, type_) in valTypes.items():
        if key in kwargs:
          val = kwargs[key]
          if isinstance(val, type_):
            initValues[key] = val
          else:
            e = typeMsg(key, val, type_)
            raise TypeError(e)

      #  Creates default values for missing arguments
      for (key, defVal) in defVals.items():
        if key not in initValues:
          initValues[key] = defVal(self)

      #  Assigns values
      for (key, value) in initValues.items():
        if value is None:
          setattr(self, key, value)

    return __init__

  @staticmethod
  def _slotsFactory(boxes: list[tuple[str, AttriBox]]) -> list[str]:
    """This factory creates the '__slots__' list which is used to restrict
    the namespace to the AttriBox instances."""
    return [box[0] for box in boxes]

  def compile(self) -> dict:
    """The namespace created by the BaseNamespace class is updated with
    the '__init__' function created by the factory function."""
    namespace = BaseNamespace.compile(self)
    oldInit = namespace.get('__init__', None)
    oldGetAttr = namespace.get('__getattr__', None)
    boxes = self._getAllBoxes()
    namespace['__field_boxes__'] = self._getFieldBoxes()
    namespace['__slots__'] = self._slotsFactory(boxes)
    if oldInit is None or oldInit is object.__init__:
      newInit = self._initFactory(boxes)
      namespace['__init__'] = newInit
    if oldGetAttr is None:
      newGetAttr = self._getattrFactory(boxes)
      namespace['__getattr__'] = newGetAttr
    return namespace

==> worktoy-0.99.71.tar.gz: worktoy-0.99.71/src/worktoy/base/_fast_space.py <==
"""FastSpace provides the namespace object class for the FastObject class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.text import monoSpace, typeMsg

try:
  from typing import Callable, Any, Self
except ImportError:
  Callable = object
  Any = object
  Self = object

from worktoy.desc import AttriBox, Bag
from worktoy.meta import OverloadSpace, Overload
from worktoy.parse import maybe, typeCast


class _ExplicitBox(AttriBox):
  """The '_ExplicitBox' class is a subclass of the 'AttriBox' class that
  allows the user to explicitly set the default value of the field."""

  __explicit_default__ = None

  def getExplicitDefault(self) -> Any:
    """This method returns the explicit default value of the field."""
    if self.__explicit_default__ is None:
      e = """The default value of the field is not explicitly set!"""
      raise ValueError(e)
    fieldCls = self.getFieldClass()
    return typeCast(self.__explicit_default__, fieldCls)

  def setExplicitDefault(self, value: Any) -> None:
    """This method sets the explicit default value of the field."""
    if self.__explicit_default__ is not None:
      e = """The default value of the field is already explicitly set!"""
      raise ValueError(e)
    fieldCls = self.getFieldClass()
    self.__explicit_default__ = typeCast(value, fieldCls)

  def __call__(self, *args, **kwargs) -> Self:
    """This method sets the default value of the field."""
    if kwargs:
      e = """The '%s' class dose not allow keyword arguments!"""
      raise TypeError(e % self.__class__.__name__)
    if not args:
      e = """The '%s' class requires one positional argument!"""
      raise TypeError(e % self.__class__.__name__)
    if len(args) - 1:
      e = """The '%s' class requires only one positional argument!"""
      raise TypeError(e % self.__class__.__name__)
    self.setExplicitDefault(args[0])
    return self

  def getDefaultFactory(self) -> Callable:
    """This method returns the default factory function."""
    explicitDefault = self.getExplicitDefault()

    def callMeMaybe(instance: object) -> Any:
      """This function returns the default value of the field."""
      return explicitDefault

    return callMeMaybe


class FastSpace(OverloadSpace):
  """EZSpace provides the namespace object class for the EZData class."""

  __field_boxes__ = None
  __inner_functions__ = None

  def _getFieldBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the field boxes."""
    return maybe(self.__field_boxes__, [])

  def _getBaseBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the base field boxes."""
    baseClasses = self.getBaseClasses()
    if not baseClasses:
      return []
    baseCls = self.getBaseClasses()[0]
    baseBoxes = getattr(baseCls, '__field_boxes__', None)
    if baseBoxes is None:
      return []
    if isinstance(baseBoxes, list):
      return baseBoxes
    e = typeMsg('baseBoxes', baseBoxes, list)
    raise TypeError(e)

  def _getAllBoxes(self, ) -> list[tuple[str, AttriBox]]:
    """This method returns all the field boxes."""
    baseBoxes = self._getBaseBoxes()
    fieldBoxes = self._getFieldBoxes()
    return [*baseBoxes, *fieldBoxes]

  def _addFieldBox(self, key: str, box: AttriBox) -> None:
    """This method adds a field box to the namespace."""
    boxes = self._getFieldBoxes()
    self.__field_boxes__ = [*boxes, (key, box)]

  def __setitem__(self, key: str, value: object) -> None:
    """This method sets the key, value pair in the namespace."""
    if isinstance(value, AttriBox):
      return self._addFieldBox(key, value)
    if callable(value) or isinstance(value, Overload):
      return OverloadSpace.__setitem__(self, key, value)
    if self.isSpecialKey(key):
      return dict.__setitem__(self, key, value)
    return self.__setitem__(key, _ExplicitBox[type(value)](value))

  @staticmethod
  def _getattrFactory(boxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__getattr__' method which automatically
    retrieves the AttriBox instances."""

    keys = [key for (key, box) in boxes]
    defGet = {key: box.getDefaultFactory() for (key, box) in boxes}

    def __getattr__(self, key: str) -> object:
      """This automatically generated '__getattr__' method retrieves the
      AttriBox instances."""
      if key in defGet:
        setattr(self, key, defGet[key](self))
      return object.__getattribute__(self, key)

    return __getattr__

  @staticmethod
  def _initFactory(attriBoxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__init__' method which automatically
    populates the AttriBox instances."""

    keys = [key for (key, box) in attriBoxes]
    defVals = {key: box.getDefaultFactory() for (key, box) in attriBoxes}
    valTypes = {key: box.getFieldClass() for (key, box) in attriBoxes}

    def __init__(self, *args, **kwargs) -> None:
      """This automatically generated '__init__' method populates the
      AttriBox instances."""
      initValues = dict()  # Creates temporary dictionary for initial values

      #  Retrieves values from positional arguments
      for (i, (key, type_)) in enumerate(valTypes.items()):
        if len(args) > i:
          if isinstance(args[i], type_):
            initValues[key] = args[i]
          else:
            e = typeMsg(key, args[i], type_)
            raise TypeError(e)

      #  Retrieves values from keyword arguments
      for (key, type_) in valTypes.items():
        if key in kwargs:
          val = kwargs[key]
          if isinstance(val, type_):
            initValues[key] = val
          else:
            e = typeMsg(key, val, type_)
            raise TypeError(e)

      #  Creates default values for missing arguments
      for (key, defVal) in defVals.items():
        if key not in initValues:
          initValues[key] = defVal(self)

      #  Assigns values
      for (key, value) in initValues.items():
        if value is None:
          setattr(self, key, value)

    return __init__

  @staticmethod
  def _slotsFactory(boxes: list[tuple[str, AttriBox]]) -> list[str]:
    """This factory creates the '__slots__' list which is used to restrict
    the namespace to the AttriBox instances."""
    return [box[0] for box in boxes]

  def compile(self) -> dict:
    """The namespace created by the BaseNamespace class is updated with
    the '__init__' function created by the factory function."""
    namespace = OverloadSpace.compile(self)
    oldInit = namespace.get('__init__', None)
    oldGetAttr = namespace.get('__getattr__', None)
    boxes = self._getAllBoxes()
    namespace['__field_boxes__'] = self._getFieldBoxes()
    namespace['__slots__'] = self._slotsFactory(boxes)
    if oldInit is None or oldInit is object.__init__:
      newInit = self._initFactory(boxes)
      namespace['__init__'] = newInit
    if oldGetAttr is None:
      newGetAttr = self._getattrFactory(boxes)
      namespace['__getattr__'] = newGetAttr
    return namespace
